const getFunctionBody = funcString => {
  return funcString.slice(funcString.indexOf("{") + 1, funcString.lastIndexOf("}"));
};

function getArgs(funcString) {
  return funcString.replace(/[/][/].*$/gm, "").replace(/\s+/g, "").replace(/[/][*][^/*]*[*][/]/g, "").split("){", 1)[0].replace(/^[^(]*[(]/, "").replace(/=[^,]+/g, "").split(",").filter(Boolean);
}

function setupTransformFunction(func, argName, returnName) {
  const funcString = func + "";
  const [valueArg] = getArgs(funcString);
  const val = getFunctionBody(funcString).replace(new RegExp(`\\b${valueArg}\\b`, "g"), argName).replace("return", returnName + " =");
  return val;
}

const createUintParser = (context, bytes, endianness) => {
  const keys = [...new Array(bytes).keys()];

  if (endianness === "big") {
    keys.reverse();
  }

  const valueVariable = context.availableVariables.shift();
  const get = context.output + " = " + keys.map(i => `data[offset++]${i === 0 ? "" : ` * 2 ** ${8 * i}`}`).join(" + ") + ";";
  const put = `
let ${valueVariable} = ${context.input};
${keys.map(i => `data[offset${i > 0 ? ` + ${i}` : ""}] = ${valueVariable};
${valueVariable} = ${valueVariable} >>> 8;`).join("\r\n")}
offset += ${bytes};`;
  return {
    get,
    put
  };
};

const uint8 = context => createUintParser(context, 1, "big");
const uint16be = context => createUintParser(context, 2, "big");
const uint16le = context => createUintParser(context, 2, "little");
const uint32be = context => createUintParser(context, 4, "big");
const uint32le = context => createUintParser(context, 4, "little");
const int8 = context => {
  const uintParser = createUintParser(context, 1, "big");
  const valueVariable = context.availableVariables.shift();
  return {
    get: `
  const ${valueVariable} = data[offset++];
  ${context.output} = ${valueVariable} | (${valueVariable} & 2 ** 7) * 0x1fffffe;`,
    put: uintParser.put
  };
};
const int16be = context => {
  const uintParser = createUintParser(context, 2, "big");
  const valueVariable = context.availableVariables.shift();
  return {
    get: `
const ${valueVariable} = data[offset++] * 2 ** 8 + data[offset++];
${context.output} = ${valueVariable} | (${valueVariable} & 2 ** 15) * 0x1fffe;`,
    put: uintParser.put
  };
};
const int16le = context => {
  const uintParser = createUintParser(context, 2, "little");
  const valueVariable = context.availableVariables.shift();
  return {
    get: `
const ${valueVariable} = data[offset++] + data[offset++] * 2 ** 8;
${context.output} = ${valueVariable} | (${valueVariable} & 2 ** 15) * 0x1fffe;`,
    put: uintParser.put
  };
};
const int32be = context => {
  const uintParser = createUintParser(context, 4, "big");
  return {
    get: `${context.output} = (data[offset++] <<  24) + data[offset++] * 2 ** 16 + data[offset++] * 2 ** 8 + data[offset++];`,
    put: uintParser.put
  };
};
const int32le = context => {
  const uintParser = createUintParser(context, 4, "little");
  return {
    get: `${context.output} = data[offset++] + data[offset++] * 2 ** 8 + data[offset++] * 2 ** 16 + (data[offset++] <<  24);`,
    put: uintParser.put
  };
};
const boolean = context => ({
  get: `${context.output} = !!data[offset++];`,
  put: `data[offset++] = ${context.input} ? 1 : 0;`
});
function skip(bytesToSkip = 1) {
  const body = `offset += ${bytesToSkip};`;
  return () => ({
    get: body,
    put: body
  });
}
function writeOnly(writer, value) {
  return context => {
    const throwawayVariable = context.availableVariables.shift();
    const writerParser = writer({ ...context,
      input: throwawayVariable,
      output: throwawayVariable
    });
    return {
      get: `
let ${throwawayVariable};
${writerParser.get}
`,
      put: `
const ${throwawayVariable} = ${JSON.stringify(value)};
${writerParser.put}
`
    };
  };
}

const addToContext = (context, val) => ({ ...context,
  output: context.output + val,
  input: context.input + val
});

const array = opts => context => {
  const itemVariable = context.availableVariables.shift();
  const itemParser = opts.of({ ...context,
    input: itemVariable,
    output: itemVariable
  });
  const arrayVariable = context.availableVariables.shift();
  const lengthVariable = context.availableVariables.shift();
  let lengthParser;

  if (typeof opts.length !== "number") {
    lengthParser = opts.of({ ...context,
      input: lengthVariable,
      output: lengthVariable
    });
  }

  const get = `
const ${arrayVariable} = [];
${lengthParser ? lengthParser.get : `const ${lengthVariable} = ${opts.length};`}
for (let i = 0; i < ${lengthVariable}; i++) {
  let ${lengthVariable};
  ${itemParser.get}
  ${arrayVariable}.push(${itemVariable});
}
${context.output} = ${arrayVariable}`;
  const put = `
${lengthParser ? `
const ${lengthVariable} = ${context.input}.length;
${lengthParser.put}
` : ""}
for (const ${itemVariable} of ${context.input}) {
  ${itemParser.put}
}
`;
  return {
    get,
    put
  };
};
const string = (length, encoding = "utf8") => context => {
  const lengthVariable = context.availableVariables.shift();
  const outputTempVariable = context.availableVariables.shift();
  const lengthContext = { ...context,
    input: lengthVariable,
    output: lengthVariable
  };
  const {
    get: getLength,
    put: setLength
  } = length(lengthContext);
  const get = `
let ${lengthVariable};
${getLength}
${context.output}
  = new TextDecoder("${encoding}").decode(data.slice(offset, offset += ${lengthVariable}));`;
  const put = `
const ${lengthVariable} = ${context.input}.length;
${setLength}
const ${outputTempVariable}
  = Buffer.from(${context.input}, "ascii");

data.set(${outputTempVariable}, offset);
offset += ${context.input}.length;
  `;
  return {
    get,
    put
  };
};
const transform = (v, to, from) => context => {
  const valueVariable = context.availableVariables.shift();
  const valueParser = v({ ...context,
    input: valueVariable,
    output: valueVariable
  });
  const get = `
let ${valueVariable};
${valueParser.get}
${setupTransformFunction(to, valueVariable, context.output)}
`;
  const put = `
let ${valueVariable};
${setupTransformFunction(from, context.input, valueVariable)}
${valueParser.put}
`;
  return {
    get,
    put
  };
};
function field(name, value) {
  return context => typeof value === "function" ? value(addToContext(context, "." + name)) : {
    get: `${context.output}.${name} = ${JSON.stringify(value)};`,
    put: ``
  };
}
function record(...fields) {
  return context => fields.reduce((result, f) => {
    const field = f(context);
    return {
      get: result.get + field.get + "\r\n",
      put: result.put + field.put + "\r\n"
    };
  }, {
    get: `${context.output} = {};`,
    put: ""
  });
}
function compile(objectParser) {
  const builtParser = objectParser({
    availableVariables: Array.from(Array(100).keys()).map(i => `tmp_${i}`),
    input: "obj",
    output: "obj"
  });
  let parse;
  let generateTo;
  const parseFuncBody = `
  let offset = 0;
  let obj;
  ${builtParser.get}
  return obj;
    `;

  try {
    parse = new Function("data", parseFuncBody);
  } catch (ex) {
    console.info(parseFuncBody);
    throw ex;
  }

  const generateFuncBody = `
${builtParser.put}
return offset;
  `;

  try {
    generateTo = new Function("offset", "data", "obj", generateFuncBody);
  } catch (ex) {
    console.info(parseFuncBody);
    throw ex;
  }

  return {
    parse,
    generate: obj => {
      let offset = 0;
      let data = new Uint8Array(2048);
      offset = generateTo(offset, data, obj);
      return data.subarray(0, offset);
    },
    generateTo
  };
}

export { uint8, uint16be, uint16le, uint32be, uint32le, int8, int16be, int16le, int32be, int32le, boolean, skip, writeOnly, array, string, transform, field, record, compile };
